from __future__ import annotations

import io
import json
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from arakawa.client import log

from .exceptions import NotebookError

if TYPE_CHECKING:
    from IPython.core.interactiveshell import InteractiveShell


_env = None


def get_environment() -> PythonEnvironment:
    """Returns the current IPython environment"""
    global _env
    if _env is None:
        _env = _get_environment()
        log.info("Detected IPython environment: %s", _env.name)
    return _env


def _get_ipython() -> InteractiveShell | None:
    try:
        return get_ipython()  # type: ignore
    except NameError:
        from IPython import get_ipython

        return get_ipython()


def is_zqm_interactive_shell() -> bool:
    """Check if we are running in a ZMQ interactive shell"""
    return _get_ipython().__class__.__name__ == "ZMQInteractiveShell"


def is_terminal_interactive_shell() -> bool:
    """Check if we are running in a terminal interactive shell"""
    return _get_ipython().__class__.__name__ == "TerminalInteractiveShell"


def get_ipython_user_ns() -> dict:
    return _get_ipython().user_ns  # type: ignore


def _get_environment() -> PythonEnvironment:  # noqa: C901
    """Determines the current IPython environment and returns an instance of the appropriate class"""

    # TODO: change this to be a list, and put the logic in a `.supported()` method
    if "google.colab" in sys.modules:
        return ColabEnvironment()
    if {"CODESPACES", "JUPYTERLAB_PATH", "VSCODE_CWD"}.issubset(set(os.environ)):
        return CodespacesVSCodeJupyterEnvironment()
    if {"CODESPACES", "JUPYTERLAB_PATH", "JPY_PARENT_PID"}.issubset(set(os.environ)):
        return CodespacesJupyterLabEnvironment()
    if is_zqm_interactive_shell() and "PAPERMILL_OUTPUT_PATH" in get_ipython_user_ns():
        return PapermillEnvironment()
    if is_zqm_interactive_shell() and "VSCODE_PID" in os.environ:
        return VSCodeJupyterEnvironment()
    if is_zqm_interactive_shell() and {"JPY_SESSION_NAME", "JPY_PARENT_PID"}.issubset(
        set(os.environ)
    ):
        return JupyterLabEnvironment()
    if is_zqm_interactive_shell() and "JPY_PARENT_PID" in os.environ:
        return JupyterNotebookEnvironment()
    if is_zqm_interactive_shell():
        return UnsupportedNotebookEnvironment()
    if "PYCHARM_HOSTED" in os.environ:
        return PyCharmEnvironment()
    if "VSCODE_PID" in os.environ:
        return VSCodeEnvironment()
    if is_terminal_interactive_shell():
        return IPythonTerminalEnvironment()
    return UnrecognizedEnvironment()


class PythonEnvironment:
    name: str = "Python Environment"
    can_open_links_from_python: bool = True
    is_notebook_environment: bool = False
    support_rich_display: bool = True
    supports_ipywidgets: bool = False

    get_ipython = staticmethod(_get_ipython)

    def get_notebook_json(self) -> dict:
        try:
            path = self._get_notebook_path()
            with open(path, encoding="utf-8") as f:
                notebook_json = json.load(f)
        except FileNotFoundError as e:
            raise FileNotFoundError(
                "Notebook not found. This command must be executed from within a Jupyter notebook environment."
            ) from e
        return notebook_json

    def _get_notebook_path(self) -> Path:
        """Get the path to the current notebook"""
        raise FileNotFoundError("Could not discover notebook path")


class UnrecognizedEnvironment(PythonEnvironment):
    name = "Unrecognized Environment"
    is_notebook_environment = False
    can_open_links_from_python = False


class IPythonTerminalEnvironment(PythonEnvironment):
    name = "IPython Terminal Environment"


class IPythonZMQEnvironment(IPythonTerminalEnvironment):
    name = "IPython ZMQ Environment"
    is_notebook_environment = True

    def _get_notebook_path(self) -> Path:
        # added in ipykernel v6.21
        # https://github.com/ipython/ipykernel/pull/1078
        user_ns = get_ipython_user_ns()

        # If it's an actual path, it'll always be absolute
        if (path := user_ns.get("__session__", None)) and path.startswith("/"):
            return Path(path)

        return super()._get_notebook_path()


class UnsupportedNotebookEnvironment(IPythonZMQEnvironment):
    name = "Unsupported Notebook Environment"


class PyCharmEnvironment(PythonEnvironment):
    name = "PyCharm Environment"
    support_rich_display = False


class VSCodeEnvironment(PythonEnvironment):
    name = "VSCode Environment"


class JupyterEnvironment(IPythonZMQEnvironment):
    name = "Jupyter Environment"
    supports_ipywidgets = True

    def _get_notebook_path(self):
        """Get the path for the current Jupyter notebook"""

        # ipynbname plays with the session API, so see if ipykernel gave us a better option
        try:
            return super()._get_notebook_path()
        except FileNotFoundError:
            import ipynbname

            try:
                return ipynbname.path()
            except IndexError as e:
                raise e


class JupyterLabEnvironment(JupyterEnvironment):
    name = "JupyterLab Environment"


class JupyterNotebookEnvironment(JupyterEnvironment):
    name = "Jupyter Notebook Environment"


class VSCodeJupyterEnvironment(JupyterEnvironment):
    name = "VSCode Jupyter Environment"

    # Attempting to use IPyWidgetsControllerUI from within VSCode Jupyter extension
    # seems extremely flakey:
    # - server often fails to start or stop
    # - once it has failed once, it can require restarting VSCode to fix things.
    # So it's better to fall back to something that seems to work better.
    #
    supports_ipywidgets = False

    def _get_notebook_path(self) -> Path:
        user_ns = get_ipython_user_ns()
        if vsc_path := user_ns.get("__vsc_ipynb_file__", None):
            return Path(vsc_path)

        # VSCode path name in sesssion may be suffixed with -jvsc-[identifier]
        # https://github.com/microsoft/vscode-jupyter/blob/113c3e54ac1d3cb81ab6473d1a5fa4a20cce4755/src/kernels/helpers.ts#L149-L168
        sessioned_nb_path = super()._get_notebook_path()
        suffix = sessioned_nb_path.suffix
        sans_suffix = sessioned_nb_path.with_suffix("")
        return Path(str(sans_suffix).split("-jvsc-")[0]).with_suffix(suffix)


class ColabEnvironment(IPythonZMQEnvironment):
    name = "Google Colab Environment"

    def get_notebook_json(self) -> dict:
        """Get the JSON for the current Colab notebook"""
        import ipynbname
        from google.colab import auth
        from googleapiclient.discovery import build
        from googleapiclient.http import MediaIoBaseDownload

        # Get the notebook's Google Drive file_id
        file_id = ipynbname.name().replace("fileId=", "")

        try:
            auth.authenticate_user()
        except Exception as e:
            raise NotebookError(
                "Google Drive authentication failed. Please allow this notebook to access your Google Drive."
            ) from e

        drive_service = build("drive", "v3")

        request = drive_service.files().get_media(fileId=file_id)
        downloaded = io.BytesIO()
        downloader = MediaIoBaseDownload(downloaded, request)
        done = False
        while done is False:
            # _ is a placeholder for a progress object that we ignore.
            # (Our file is small, so we skip reporting progress.)
            _, done = downloader.next_chunk()

        downloaded.seek(0)
        return json.loads(downloaded.read().decode("utf-8"))


class CodespacesVSCodeJupyterEnvironment(VSCodeJupyterEnvironment):
    name = "Codespaces VSCode Jupyter Environment"


class CodespacesJupyterLabEnvironment(JupyterLabEnvironment):
    name = "Codespaces JupyterLab Environment"
    can_open_links_from_python = False


class PapermillEnvironment(IPythonZMQEnvironment):
    name = "Papermill Environment"
    support_rich_display = False

    def _get_notebook_path(self) -> Path:
        user_ns = get_ipython_user_ns()
        if path := user_ns.get("PAPERMILL_OUTPUT_PATH", None):
            return Path(path)
        return super()._get_notebook_path()



"""
# TODO - optimise import handling here
"""
# ruff: noqa: FA100

import json
import pickle
from contextlib import suppress
from io import TextIOWrapper
from typing import Any, Union

import pandas as pd
from altair.utils import SchemaBase
from multimethod import multimethod
from pandas.io.formats.style import Styler

from arakawa import optional_libs as opt
from arakawa.client import log
from arakawa.common import ArrowFormat
from arakawa.exceptions import ARError

from .xml_visitor import AssetMeta


class DPTextIOWrapper(TextIOWrapper):
    """Custom IO Wrapper that detaches before closing - see https://bugs.python.org/issue21363"""

    def __init__(self, f, *a, **kw):
        super().__init__(f, *a, encoding="utf-8", **kw)

    def __del__(self):
        # don't close the underlying stream
        with suppress(Exception):
            self.flush()
        with suppress(Exception):
            self.detach()


class AttachmentWriter:
    # pickle
    @multimethod
    def get_meta(self, x: Any) -> AssetMeta:
        return AssetMeta(ext=".pkl", mime="application/vnd.pickle+binary")

    @multimethod
    def get_meta(self, x: str) -> AssetMeta:
        return AssetMeta(ext=".json", mime="application/json")

    @multimethod
    def write_file(self, x: Any, f) -> None:
        pickle.dump(x, f)

    @multimethod
    def write_file(self, x: str, f) -> None:
        out: str = json.dumps(json.loads(x))
        f.write(out.encode())


class DataTableWriter:
    @multimethod
    def get_meta(self, x: pd.DataFrame) -> AssetMeta:
        return AssetMeta(mime=ArrowFormat.content_type, ext=ArrowFormat.ext)

    @multimethod
    def write_file(self, x: pd.DataFrame, f) -> None:
        if x.size == 0:
            raise ARError("Empty DataFrame provided")
        # process_df called in Arrow.save_file
        ArrowFormat.save_file(f, x)


class HTMLTableWriter:
    @multimethod
    def get_meta(self, x: Union[pd.DataFrame, Styler]) -> AssetMeta:
        return AssetMeta(mime="application/vnd.arakawa.table+html", ext=".tbl.html")

    @multimethod
    def write_file(self, x: pd.DataFrame, f) -> None:
        self._check(x)
        out = x.to_html().encode()
        f.write(out)

    @multimethod
    def write_file(self, x: Styler, f) -> None:
        self._check(x.data)
        out = x.to_html().encode()
        f.write(out)

    def _check(self, df: pd.DataFrame) -> None:
        n_cells = df.shape[0] * df.shape[1]
        if n_cells > 500:
            log.warning(
                "Table is over recommended size, consider using ar.DataTable instead or aggregating the df first"
            )


class PlotWriter:
    obj_type: Any

    # Altair (always installed)
    @multimethod
    def get_meta(self, x: SchemaBase) -> AssetMeta:
        return AssetMeta(mime="application/vnd.vegalite.v5+json", ext=".vl.json")

    @multimethod
    def write_file(self, x: SchemaBase, f) -> None:
        json.dump(x.to_dict(), DPTextIOWrapper(f))

    if opt.HAVE_FOLIUM:

        @multimethod
        def get_meta(self, x: opt.Map) -> AssetMeta:
            return AssetMeta(mime="application/vnd.folium+html", ext=".fl.html")

        @multimethod
        def write_file(self, x: opt.Map, f) -> None:
            html: str = x.get_root().render()
            f.write(html.encode())

    if opt.HAVE_BOKEH:

        @multimethod
        def get_meta(self, x: Union[opt.BFigure, opt.BLayout]) -> AssetMeta:
            return AssetMeta(mime="application/vnd.bokeh.show+json", ext=".bokeh.json")

        @multimethod
        def write_file(self, x: Union[opt.BFigure, opt.BLayout], f):
            from bokeh.embed import json_item

            json.dump(json_item(x), DPTextIOWrapper(f))

    if opt.HAVE_PLOTLY:

        @multimethod
        def get_meta(self, x: opt.PFigure) -> AssetMeta:
            return AssetMeta(mime="application/vnd.plotly.v1+json", ext=".pl.json")

        @multimethod
        def write_file(self, x: opt.PFigure, f):
            json.dump(x.to_json(), DPTextIOWrapper(f))

    if opt.HAVE_MATPLOTLIB:

        @multimethod
        def get_meta(self, x: Union[opt.Axes, opt.Figure, opt.ndarray]) -> AssetMeta:
            return AssetMeta(mime="image/svg+xml", ext=".svg")

        @multimethod
        def write_file(self, x: opt.Figure, f) -> None:
            x.savefig(DPTextIOWrapper(f), format="svg", bbox_inches="tight")

        @multimethod
        def write_file(self, x: opt.Axes, f) -> None:
            self.write_file(x.get_figure(), f)

        @multimethod
        def write_file(self, x: opt.ndarray, f) -> None:
            fig = x.flatten()[0].get_figure()
            self.write_file(fig, f)



from __future__ import annotations

import re
import textwrap
from collections import deque
from pathlib import Path

from dominate.dom_tag import dom_tag

from arakawa.common.ar_types import NPath
from arakawa.common.utils import utf_read_text
from arakawa.common.viewxml_utils import get_embed_url
from arakawa.exceptions import ARError

from .base import BlockId, BlockOrPrimitive, DataBlock, wrap_block
from .layout import Group


class EmbeddedTextBlock(DataBlock):
    """
    Abstract Block for embedded text formats that are stored directly in the
    document (rather than external references)
    """

    content: str

    def __init__(self, content: str, name: BlockId | None = None, **kwargs):
        super().__init__(name, **kwargs)
        self.content = content.strip()


class Text(EmbeddedTextBlock):
    """
    You can add short or long-form Markdown content to your app with the `Text` block.

    !!! info
        Markdown is a lightweight markup language that allows you to include formatted text in your app, and can be accessed through `ar.Text`, or by passing in a string directly.&#x20;

        Check [here](https://github.com/adam-p/markdown-here/wiki/Markdown-Cheatsheet) for more information on how to format your text with markdown.
    """

    _tag = "Text"

    def __init__(
        self,
        text: str | None = None,
        file: NPath | None = None,
        name: BlockId | None = None,
        label: str | None = None,
    ):
        """
        Args:
            text: The markdown formatted text, use triple-quotes, (`\"\"\"# My Title\"\"\"`) to create multi-line markdown text
            file: Path to a file containing markdown text
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)

        !!! note
            File encodings are auto-detected, if this fails please read the file manually with an explicit encoding and use the text parameter on ar.Attachment
        """
        if text:
            text = textwrap.dedent(text).strip()

        content: str | None = None
        if file:
            content = utf_read_text(Path(file).expanduser())

        content = text or content
        assert content

        super().__init__(content=content, name=name, label=label)

    def format(self, *args: BlockOrPrimitive, **kwargs: BlockOrPrimitive) -> Group:
        """
        Format the markdown text template, using the supplied context to insert blocks into `{{}}` markers in the template.

        `{}` markers can be empty, hence positional, or have a name, e.g. `{{plot}}`, which is used to lookup the value from the keyword context.

        Args:
            *args: positional template context arguments
            **kwargs: keyword template context arguments

        !!! tip
            Either Python objects, e.g. dataframes, and plots, or Arakawa blocks as context

        Returns:
            An Arakawa Group object containing the list of text and embedded objects
        """

        splits = re.split(r"\{\{(\w*)\}\}", self.content)
        deque_args = deque(args)
        blocks = []

        for i, x in enumerate(splits):
            is_block = bool(i % 2)

            if is_block:
                try:
                    if x:
                        blocks.append(wrap_block(kwargs[x]))
                    else:
                        blocks.append(wrap_block(deque_args.popleft()))
                except (IndexError, KeyError) as e:
                    raise ARError(
                        f"Unknown/missing object '{x}' referenced in Markdown format"
                    ) from e

            else:
                x = x.strip()
                if x:
                    blocks.append(Text(x))

        return Group(blocks=blocks)


class Code(EmbeddedTextBlock):
    """
    The code block allows you to embed syntax-highlighted source code into your app.

    !!! note
        This block currently supports Python and JavaScript.
    """

    _tag = "Code"

    def __init__(
        self,
        code: str,
        language: str = "python",
        caption: str | None = None,
        name: BlockId | None = None,
        label: str | None = None,
    ):
        """
        Args:
            code: The source code
            language: The language of the code, most common languages are supported (optional - defaults to Python)
            caption: A caption to display below the Code (optional)
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)
        """
        super().__init__(
            content=code, language=language, caption=caption, name=name, label=label
        )


class HTML(EmbeddedTextBlock):
    """
    The HTML block allows you to add raw HTML to your app,  allowing for highly customized components, such as your company's brand, logo, and more.

    !!! info
        The HTML block is sandboxed and cannot execute JavaScript.
    """

    _tag = "HTML"

    def __init__(
        self, html: str | dom_tag, name: BlockId | None = None, label: str | None = None
    ):
        """
        Args:
            html: The HTML fragment to embed - can be a string or a [dominate](https://github.com/Knio/dominate/) tag
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)
        """
        super().__init__(content=str(html), name=name, label=label)


class Formula(EmbeddedTextBlock):
    """
    The formula block allows you easily to add [_LaTeX_](https://en.wikipedia.org/wiki/LaTeX)-formatted equations to your app, with an optional caption.

    !!! tip
        A brief intro into _LaTeX_ formulas can be found [here](https://en.wikibooks.org/wiki/LaTeX/Mathematics).
    """

    _tag = "Formula"

    def __init__(
        self,
        formula: str,
        caption: str | None = None,
        name: BlockId | None = None,
        label: str | None = None,
    ):
        r"""
        Args:
            formula: The formula to embed, using LaTeX format (use raw strings)
            caption: A caption to display below the Formula (optional)
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)

        !!! note
            LaTeX commonly uses special characters, hence prefix your formulas with `r` to make them raw strings, e.g. `r"\frac{1}{\sqrt{x^2 + 1}}"`

        Under the hood we use MathJAX to render the equations in the browser and not a full TeX engine. This means that some of your TeX input may not be rendered correctly on our system - read the MathJAX documentation for more info.
        """
        super().__init__(content=formula, caption=caption, name=name, label=label)


class Embed(EmbeddedTextBlock):
    """
    The Embed block lets you embed content from other platforms e.g. Youtube, Spotify.

    !!! tip
        If you're trying to embed an `iframe`, you can wrap it in an `HTML` block.
    """

    _tag = "Embed"

    def __init__(
        self,
        url: str,
        width: int = 960,
        height: int = 540,
        name: BlockId | None = None,
        label: str | None = None,
    ):
        """
        Args:
            url: The URL of the resource to be embedded
            width: The width of the embedded object (optional)
            height: The height of the embedded object (optional)
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)
        """

        result = get_embed_url(url, width=width, height=height)

        # if "html" not in result:
        #     raise DPClientError(f"Can't embed result from provider for URL '{url}'")
        super().__init__(
            content=result.html,
            name=name,
            label=label,
            url=url,
            title=result.title,
            provider_name=result.provider,
        )



from .asset import Attachment, DataTable, Media, Plot, Table  # noqa: F401
from .base import (  # noqa: F401
    BaseBlock,
    BlockList,
    BlockOrPrimitive,
    DataBlock,
    wrap_block,
)
from .empty import Empty  # noqa: F401
from .layout import Group, Page, Select, SelectType, Toggle, VAlign  # noqa: F401
from .misc_blocks import BigNumber  # noqa: F401
from .text import HTML, Code, Embed, Formula, Text  # noqa: F401

Block = BaseBlock



import dataclasses as dc
from collections import namedtuple
from typing import TYPE_CHECKING, Any, Protocol

from lxml import etree
from lxml.builder import ElementMaker
from multimethod import DispatchError, multimethod
from typing_extensions import Self

from arakawa.blocks import BaseBlock
from arakawa.blocks.asset import AssetBlock
from arakawa.blocks.layout import ContainerBlock
from arakawa.blocks.text import EmbeddedTextBlock
from arakawa.client import log
from arakawa.common.viewxml_utils import ElementT, mk_attribs
from arakawa.exceptions import ARError
from arakawa.view.view_blocks import Blocks
from arakawa.view.visitors import ViewVisitor

if TYPE_CHECKING:
    from arakawa.processors import FileEntry, FileStore

E = ElementMaker()  # XML Tag Factory


@dc.dataclass
class XMLBuilder(ViewVisitor):
    """Convert the Blocks into an XML document"""

    store: "FileStore"
    # element: t.Optional[etree.Element] = None  # Empty Group Element?
    elements: list[ElementT] = dc.field(default_factory=list)

    def get_root(self, fragment: bool = False) -> ElementT:
        """Return the top-level ViewXML"""
        # create the top-level

        # get the top-level root
        _top_group: ElementT = self.elements.pop()
        assert _top_group.tag == "Group"
        assert not self.elements

        # create top-level structure
        return E.View(
            # E.Internal(),
            *_top_group.getchildren(),
            **mk_attribs(version="1", fragment=fragment),
        )

    @property
    def store_count(self) -> int:
        return len(self.store.files)

    def add_element(self, _: BaseBlock, e: Any) -> Self:
        """Add an element to the list of nodes at the current XML tree location"""
        self.elements.append(e)
        return self

    # xml convertors
    @multimethod
    def visit(self, b: BaseBlock) -> Self:
        """Base implementation - just created an empty tag including all the initial attributes"""
        _E = getattr(E, b._tag)  # noqa: N806
        return self.add_element(b, _E(**b._attributes))

    def _visit_subnodes(self, b: ContainerBlock) -> list[ElementT]:
        cur_elements = self.elements
        self.elements = []
        b.traverse(self)  # visit subnodes
        res = self.elements
        self.elements = cur_elements
        return res

    @multimethod
    def visit(self, b: ContainerBlock) -> Self:
        sub_elements = self._visit_subnodes(b)
        # build the element
        _E = getattr(E, b._tag)  # noqa: N806
        element = _E(*sub_elements, **b._attributes)
        return self.add_element(b, element)

    @multimethod
    def visit(self, b: Blocks) -> Self:
        sub_elements = self._visit_subnodes(b)

        # Blocks are converted to Group internally
        if label := b._attributes.get("label"):
            log.info(f"Found label {label} in top-level Blocks/View")
        element = E.Group(*sub_elements, columns="1", valign="top")
        return self.add_element(b, element)

    @multimethod
    def visit(self, b: EmbeddedTextBlock) -> Self:
        # NOTE - do we use etree.CDATA wrapper?
        _E = getattr(E, b._tag)  # noqa: N806
        return self.add_element(b, _E(etree.CDATA(b.content), **b._attributes))

    @multimethod
    def visit(self, b: AssetBlock):
        """Main XMl creation method - visitor method"""
        fe = self._add_asset_to_store(b)

        _E = getattr(E, b._tag)  # noqa: N806

        e: etree._Element = _E(
            type=fe.mime,
            # size=conv_attrib(fe.size),
            # hash=fe.hash,
            **{**b._attributes, **b.get_file_attribs()},
            # src=f"attachment://{self.store_count}",
            src=f"ref://{fe.hash}",
        )

        if b.caption:
            e.set("caption", b.caption)
        return self.add_element(b, e)

    def _add_asset_to_store(self, b: AssetBlock) -> "FileEntry":
        """Default asset store handler that operates on native Python objects"""
        # import here as a very slow module due to nested imports
        # from .. import files

        # check if we already have stored this asset to the store
        # TODO - do we just persist the asset store across the session??
        if b._prev_entry:
            if type(b._prev_entry) == self.store.fw_klass:  # noqa: E721
                self.store.add_file(b._prev_entry)
                return b._prev_entry
            b._prev_entry = None

        if b.data is not None:
            # fe = files.add_to_store(self.data, s.store)
            try:
                writer = get_writer(b)
                meta: AssetMeta = writer.get_meta(b.data)
                fe = self.store.get_file(meta.ext, meta.mime)
                writer.write_file(b.data, fe.file)
                self.store.add_file(fe)
            except DispatchError as e:
                raise ARError(
                    f"{type(b.data).__name__} not supported for {self.__class__.__name__}"
                ) from e
        elif b.file is not None:
            fe = self.store.load_file(b.file)
        else:
            raise ARError("No asset to add")

        b._prev_entry = fe
        return fe


AssetMeta = namedtuple("AssetMeta", "ext mime")


class AssetWriterP(Protocol):
    """Implement these in any class to support asset writing
    for a particular AssetBlock"""

    def get_meta(self, x: Any) -> AssetMeta: ...

    def write_file(self, x: Any, f) -> None: ...


asset_mapping: dict[type[AssetBlock], type[AssetWriterP]] = {}


def get_writer(b: AssetBlock) -> AssetWriterP:
    import arakawa.blocks.asset as a

    from . import asset_writers as aw

    if not asset_mapping:
        asset_mapping.update(
            {
                a.Plot: aw.PlotWriter,
                a.Table: aw.HTMLTableWriter,
                a.DataTable: aw.DataTableWriter,
                a.Attachment: aw.AttachmentWriter,
            }
        )
    return asset_mapping[type(b)]()



from __future__ import annotations

import dataclasses as dc
import importlib.resources as ir
import json
import math
import re
from collections.abc import Sized
from numbers import Number
from typing import Any

from loguru import logger as log
from lxml import etree
from lxml.etree import DocumentInvalid
from lxml.etree import _Element as ElementT
from micawber import ProviderException, bootstrap_basic, bootstrap_noembed, cache

from arakawa.exceptions import ARError

from .ar_types import HTML, SSDict

local_view_resources = ir.files("arakawa.resources.view_resources")
rng_validator = etree.RelaxNG(file=str(local_view_resources / "full_schema.rng"))

ViewXML = str


def load_doc(x: str) -> ElementT:
    parser = etree.XMLParser(
        strip_cdata=False, recover=True, remove_blank_text=True, remove_comments=True
    )
    return etree.fromstring(x, parser=parser)


def is_valid_id(id: str) -> bool:
    """(cached) regex to check for a xsd:ID name"""
    return re.fullmatch(r"^[a-zA-Z_][\w.-]*$", id) is not None


def validate_view_doc(
    xml_str: str | None = None,
    xml_doc: ElementT | None = None,
    quiet: bool = False,
) -> bool:
    """Validate the model against the schema, throws an etree.DocumentInvalid if not"""
    assert xml_str or (xml_doc is not None)
    if xml_str:
        xml_doc = etree.fromstring(xml_str)

    try:
        rng_validator.assertValid(xml_doc)
        return True
    except DocumentInvalid:
        if not quiet:
            xml_str = (
                xml_str if xml_str else etree.tounicode(xml_doc, pretty_print=True)
            )
            log.error(
                f"Error validating report document:\n\n{xml_str}\n{rng_validator.error_log}\n"
            )
        raise


def conv_attrib(v: Any) -> str | None:
    """
    Convert a value to a str for use as an ElementBuilder attribute
    - also handles None to a string for optional field values
    """
    # TODO - use a proper serialisation framework here / lxml features
    if v is None:
        return v
    if isinstance(v, Sized) and len(v) == 0:
        return None
    if isinstance(v, str):
        return v
    if isinstance(v, Number) and not isinstance(v, bool):
        if math.isinf(v) and v > 0:
            return "INF"
        if math.isinf(v) and v < 0:
            return "-INF"
        if math.isnan(v):
            return "NaN"
        return str(v)
    return json.dumps(v)


def mk_attribs(**attribs: Any) -> SSDict:
    """convert attributes, dropping None and empty values"""
    return {
        str(k): v1 for (k, v) in attribs.items() if (v1 := conv_attrib(v)) is not None
    }


#####################################################################
# Embed Asset Helpers
providers = bootstrap_basic(cache=cache.Cache())


@dc.dataclass(frozen=True)
class Embedded:
    html: HTML
    title: str
    provider: str


def get_embed_url(url: str, width: int = 960, height: int = 540) -> Embedded:
    """Return html for an embeddable URL"""
    try:
        r = providers.request(url, maxwidth=width, maxheight=height)
    except ProviderException:
        # add NoEmbed to the list and try again
        try:
            log.debug("Initializing NoEmbed OEmbed provider")
            bootstrap_noembed(registry=providers)
            r = providers.request(url, maxwidth=width, maxheight=height)
        except ProviderException as e:
            raise ARError(
                f"No embed provider found for URL '{url}' - is there an active internet connection?"
            ) from e

    return Embedded(
        html=r["html"],
        title=r.get("title", "Title"),
        provider=r.get("provider_name", "Embedding"),
    )



from __future__ import annotations

from typing import Union

from .base import BlockId, DataBlock

NumberValue = Union[str, int, float]


class BigNumber(DataBlock):
    """
    A single number or change can often be the most important thing in an app.

    The `BigNumber`block allows you to present KPIs, changes, and statistics in a friendly way to your viewers.

    You can optionally set intent, and pass in numbers or text.
    """

    _tag = "BigNumber"

    def __init__(
        self,
        heading: str,
        value: NumberValue,
        change: NumberValue | None = None,
        prev_value: NumberValue | None = None,
        is_positive_intent: bool | None = None,
        is_upward_change: bool | None = None,
        name: BlockId | None = None,
        label: str | None = None,
    ):
        """
        Args:
            heading: A title that gives context to the displayed number
            value: The value of the number
            prev_value: The previous value to display as comparison (optional)
            change: The amount changed between the value and previous value (optional)
            is_positive_intent: Displays the change on a green background if `True`, and red otherwise. Follows `is_upward_change` if not set (optional)
            is_upward_change: Whether the change is upward or downward (required when `change` is set)
            name: A unique name for the block to reference when adding text or embedding (optional)
            label: A label used when displaying the block (optional)
        """
        if change:
            if is_upward_change is None:
                # We can't reliably infer the direction of change from the change string
                raise ValueError(
                    'Argument "is_upward_change" is required when "change" is set'
                )
            if is_positive_intent is None:
                # Set the intent to be the direction of change if not specified (up = green, down = red)
                is_positive_intent = is_upward_change

        super().__init__(
            heading=heading,
            value=value,
            change=change,
            prev_value=prev_value,
            is_positive_intent=bool(is_positive_intent),
            is_upward_change=bool(is_upward_change),
            name=name,
            label=label,
        )



from __future__ import annotations

import importlib.resources as ir
import json
from abc import ABC
from copy import copy
from itertools import count
from os import path as osp
from pathlib import Path
from typing import Any, BinaryIO, cast
from uuid import uuid4

from jinja2 import Environment, FileSystemLoader, Template
from jinja2.utils import htmlsafe_json_dumps
from lxml import etree

from arakawa import blocks as b
from arakawa.client.utils import display_msg, log, open_in_browser
from arakawa.common import HTML, NPath, timestamp, validate_view_doc
from arakawa.common.viewxml_utils import ElementT, local_view_resources
from arakawa.exceptions import InvalidReportError
from arakawa.view import PreProcess, XMLBuilder

from .file_store import FileEntry
from .types import BaseProcessor, Formatting


class PreProcessView(BaseProcessor):
    """Optimization to improve the layout of the view using the Block-API"""

    def __init__(self, *, is_finalized: bool = True) -> None:
        self.is_finalized = is_finalized
        super().__init__()

    def __call__(self, _: Any) -> None:
        # AST checks
        if len(self.s.blocks.blocks) == 0:
            raise InvalidReportError(
                "Empty blocks object - must contain at least one block"
            )

        # convert Page -> Select + Group
        v = copy(self.s.blocks)
        if all(isinstance(blk, b.Page) for blk in v.blocks):
            # convert to top-level Select
            v.blocks = [
                b.Select(
                    blocks=[
                        b.Group(blocks=p.blocks, label=p.title, name=p.name)
                        for p in cast(list[b.Page], v.blocks)
                    ],
                    type=b.SelectType.TABS,
                )
            ]

        # Block-API visitors
        pp = PreProcess(is_finalized=self.is_finalized)
        v.accept(pp)
        v1 = pp.root
        # v1 = copy(v)

        # update the processor state
        self.s.blocks = v1

        return


class ConvertXML(BaseProcessor):
    """Convert the View AST into an XML fragment"""

    local_post_xslt = etree.parse(str(local_view_resources / "local_post_process.xslt"))
    local_post_transform = etree.XSLT(local_post_xslt)

    def __init__(self, *, pretty_print: bool = False, fragment: bool = False) -> None:
        self.pretty_print: bool = pretty_print
        self.fragment: bool = fragment
        super().__init__()

    def __call__(self, _: Any) -> ElementT:
        initial_doc = self.convert_xml()
        transformed_doc = self.post_transforms(initial_doc)

        # convert to string
        view_xml_str: str = etree.tounicode(
            transformed_doc, pretty_print=self.pretty_print
        )
        # s1 = dc.replace(s, view_xml=view_xml_str)
        self.s.view_xml = view_xml_str

        log.debug(etree.tounicode(transformed_doc, pretty_print=True))

        # return the doc for further processing (xml str stored in state)
        return transformed_doc

    def convert_xml(self) -> ElementT:
        # create initial state
        builder_state = XMLBuilder(store=self.s.store)
        self.s.blocks.accept(builder_state)
        return builder_state.get_root(self.fragment)

    def post_transforms(self, view_doc: ElementT) -> ElementT:
        # TODO - post-xml transformations, essentially xslt / lxml-based DOM operations
        # post_process via xslt
        processed_view_doc: ElementT = self.local_post_transform(view_doc)

        # TODO - custom lxml-based transforms go here...

        # validate post all transformations
        validate_view_doc(xml_doc=processed_view_doc)
        return processed_view_doc


class PreUploadProcessor(BaseProcessor):
    def __call__(self, doc: ElementT) -> tuple[str, list[BinaryIO]]:
        """
        pre-upload pass of the XML doc so can be uploaded to DPCloud
        modifies the document based on the FileStore
        """

        # NOTE - this currently relies on all assets existing linearly in document order
        # in the asset store - if we move to a cas we will need to update the algorithm here
        # replace ref -> attachment in view
        # all blocks with a ref
        refs: list[ElementT] = doc.xpath("/View//*[@src][starts-with(@src, 'ref://')]")
        for idx, ref, f_entry in zip(count(0), refs, self.s.store.files):
            ref: ElementT
            f_entry: FileEntry
            _hash: str = ref.get("src").split("://")[1]
            ref.set("src", f"attachment://{idx}")
            assert _hash == f_entry.hash  # sanity check

        self.s.view_xml = etree.tounicode(doc)
        return (self.s.view_xml, self.s.store.file_list)


###############################################################################
# HTML Exporting Processors
class BaseExportHTML(BaseProcessor, ABC):
    """Provides shared logic for writing an app to local disk"""

    # Type is `ir.abc.Traversable` which extends `Path`,
    # but the former isn't compatible with `shutil`
    template_dir: Path = cast(Path, ir.files("arakawa.resources.html_templates"))
    template: Template

    template_name: str

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        template_loader = FileSystemLoader(cls.template_dir)
        template_env = Environment(loader=template_loader)
        cls.template = template_env.get_template(cls.template_name)

    def get_cdn(self) -> str:
        from arakawa import settings

        return settings.AR_CDN_BASE

    def _write_html_template(
        self,
        name: str,
        formatting: Formatting | None = None,
        cdn_base: str | None = None,
    ) -> tuple[str, str]:
        """Internal method to write the ViewXML and assets into a HTML container and associated files"""
        name = name or "app"
        formatting = formatting or Formatting()

        report_id: str = uuid4().hex

        # TODO - split this out?
        vs = self.s
        if vs:
            assets = vs.store.as_dict() or {}
            view_xml = vs.view_xml
        else:
            assets = {}
            view_xml = ""

        app_data = {"view_xml": view_xml, "assets": assets}
        html = self.template.render(
            # Escape JS multi-line strings
            app_data=htmlsafe_json_dumps(app_data),
            report_width_class=formatting.width.to_css(),
            report_name=name,
            report_date=timestamp(),
            css_header=formatting.to_css(),
            is_light_prose=json.dumps(formatting.light_prose),
            events=False,
            report_id=report_id,
            cdn_base=cdn_base or self.get_cdn(),
        )

        return html, report_id


class ExportBaseHTMLOnly(BaseExportHTML):
    """Export the base view used to render an App, containing no ViewXML nor Assets"""

    # TODO (JB) - Create base HTML-only template
    template_name = "local_template.html.j2"

    def __init__(
        self,
        debug: bool,
        formatting: Formatting | None = None,
    ):
        self.debug = debug
        self.formatting = formatting

    def generate_chrome(self) -> HTML:
        # TODO - this is a bit hacky
        self.s = None
        html, _ = self._write_html_template(
            "app",
            formatting=self.formatting,
        )
        return HTML(html)

    def get_cdn(self) -> str:
        return "/web-static" if self.debug else super().get_cdn()

    def __call__(self, _: Any) -> None:
        return None


class ExportHTMLInlineAssets(BaseExportHTML):
    """
    Export a view into a single HTML file containing:
    - View XML - embedded
    - Assets - embedded as b64 data-uris
    """

    template_name = "local_template.html.j2"

    def __init__(
        self,
        path: str,
        open: bool = False,
        name: str = "app",
        formatting: Formatting | None = None,
        cdn_base: str | None = None,
    ):
        self.path = path
        self.open = open
        self.name = name
        self.formatting = formatting
        self.cdn_base = cdn_base

    def __call__(self, _: Any) -> str:
        html, report_id = self._write_html_template(
            name=self.name,
            formatting=self.formatting,
            cdn_base=self.cdn_base,
        )

        Path(self.path).write_text(html, encoding="utf-8")

        display_msg(f"App saved to ./{self.path}")

        if self.open:
            path_uri = f"file://{osp.realpath(osp.expanduser(self.path))}"
            open_in_browser(path_uri)

        return report_id


class ExportHTMLFileAssets(BaseExportHTML):
    """
    Export a view into a single HTML file on disk, containing
    - View XML - embedded
    - Assets - referenced as remote resources
    """

    template_name = "local_template.html.j2"

    def __init__(
        self,
        app_dir: Path,
        name: str = "app",
        formatting: Formatting | None = None,
        cdn_base: str | None = None,
    ):
        self.app_dir = app_dir
        self.name = name
        self.formatting = formatting
        self.cdn_base = cdn_base

    def __call__(self, dest: NPath | None = None) -> Path:
        html, _ = self._write_html_template(
            name=self.name,
            formatting=self.formatting,
            cdn_base=self.cdn_base,
        )

        index_path = self.app_dir / "index.html"
        index_path.write_text(html, encoding="utf-8")
        display_msg(f"Built app in {self.app_dir}")
        return self.app_dir


class ExportHTMLStringInlineAssets(BaseExportHTML):
    """
    Export the View as an in-memory string representing a resizable HTML fragment, containing
    - View XML - embedded
    - Assets - embedded as b64 data-uris
    """

    template_name = "ipython_template.html.j2"

    def __init__(
        self,
        name: str = "Stringified App",
        formatting: Formatting | None = None,
        cdn_base: str | None = None,
    ):
        self.name = name
        self.formatting = formatting
        self.cdn_base = cdn_base

    def __call__(self, _: Any) -> HTML:
        html, _ = self._write_html_template(
            name=self.name, formatting=self.formatting, cdn_base=self.cdn_base
        )
        return HTML(html)



# Copyright 2020 StackHut Limited (trading as Datapane)
# SPDX-License-Identifier: Apache-2.0
import importlib.metadata

__version__ = importlib.metadata.version(__name__)

# Public API re-exports
from .blocks import (  # noqa: F401
    HTML,
    Attachment,
    BigNumber,
    Block,
    Code,
    DataTable,
    Embed,
    Empty,
    Formula,
    Group,
    Media,
    Page,
    Plot,
    Select,
    SelectType,
    Table,
    Text,
    Toggle,
    VAlign,
    wrap_block,
)
from .client import enable_logging, print_debug_info  # noqa: F401
from .exceptions import ARError  # noqa: F401
from .processors import (  # noqa: F401
    FontChoice,
    Formatting,
    TextAlignment,
    Width,
    build_report,
    save_report,
    stringify_report,
)
from .view import Blocks, Report, View  # noqa: F401

from . import builtins  # noqa: F401 # isort: skip

X = wrap_block



from __future__ import annotations

import importlib.resources as ir
import locale
import mimetypes
import re
import sys
from pathlib import Path

import chardet
from chardet.universaldetector import UniversalDetector
from loguru import logger as log

from .ar_types import MIME

################################################################################
# CONSTANTS
ON_WINDOWS = sys.platform == "win32"

################################################################################
# MIME-type handling
mimetypes.init(files=[str(ir.files("arakawa.resources") / "mime.types")])

# TODO - hardcode as temporary fix until mimetypes double extension issue is sorted
_double_ext_map = {
    ".vl.json": "application/vnd.vegalite.v5+json",
    ".vl2.json": "application/vnd.vegalite.v2+json",
    ".vl3.json": "application/vnd.vegalite.v3+json",
    ".vl4.json": "application/vnd.vegalite.v4+json",
    ".vl5.json": "application/vnd.vegalite.v5+json",
    ".bokeh.json": "application/vnd.bokeh.show+json",
    ".pl.json": "application/vnd.plotly.v1+json",
    ".fl.html": "application/vnd.folium+html",
    ".tbl.html": "application/vnd.arakawa.table+html",
    ".tar.gz": "application/x-tgz",
}
double_ext_map: dict[str, MIME] = {k: MIME(v) for k, v in _double_ext_map.items()}


def guess_type(filename: Path) -> MIME:
    ext = "".join(filename.suffixes)
    if ext in double_ext_map:
        return double_ext_map[ext]
    mtype, _ = mimetypes.guess_type(str(filename))
    assert mtype
    return MIME(mtype or "application/octet-stream")


def guess_encoding(fn: str) -> str:
    with open(fn, "rb") as f:
        detector = UniversalDetector()
        for line in f.readlines():
            detector.feed(line)
            if detector.done:
                break
        detector.close()
    return detector.result["encoding"]  # type: ignore


def utf_read_text(file: Path) -> str:
    """Encoding-aware text reader
    - handles cases like on Windows where a file is UTF-8, but default locale is windows-1252
    """
    if ON_WINDOWS:
        f_bytes = file.read_bytes()
        f_enc: str = chardet.detect(f_bytes)["encoding"]  # type: ignore
        # NOTE - can just special case utf-8 files here?
        def_enc = locale.getpreferredencoding()
        log.debug(f"Default encoding is {def_enc}, file encoded as {f_enc}")
        if def_enc.upper() != f_enc.upper():
            log.warning(f"Text file {file} encoded as {f_enc}, auto-converting")
        return f_bytes.decode(encoding=f_enc)
    # for linux/macOS assume utf-8
    return file.read_text()


def dict_drop_empty(xs: dict | None = None, none_only: bool = False, **kwargs) -> dict:
    """Return a new dict with the empty/falsey values removed"""
    xs = {**(xs or {}), **kwargs}

    if none_only:
        return {k: v for (k, v) in xs.items() if v is not None}
    return {k: v for (k, v) in xs.items() if v or isinstance(v, bool)}


def should_compress_mime_type_for_upload(mime_type: str) -> bool:
    # This strategy is based on:
    # - looking at mime type databases used by `mimetypes` module
    # - our custom mime types in double_ext_map
    # - some other online sources that capture real-world usage:
    #   - https://letstalkaboutwebperf.com/en/gzip-brotli-server-config/
    #   - https://github.com/h5bp/server-configs-nginx/blob/main/mime.types
    return any(
        pattern.search(mime_type) for pattern in _SHOULD_COMPRESS_MIME_TYPE_REGEXPS
    )


_SHOULD_COMPRESS_MIME_TYPE_REGEXPS = [
    re.compile(p)
    for p in [
        r"^text/",
        r"\+json$",
        r"\+xml$",
        r"\+html$",
        r"^application/json$",
        r"^application/vnd\.pickle\+binary$",
        r"^application/vnd\.apache\.arrow\+binary$",
        r"^application/xml$",
    ]
]



"""
Arakawa Processors

API for processing Views, e.g. rendering it locally and publishing to a remote server
"""

from __future__ import annotations

import os
from pathlib import Path
from shutil import rmtree

from arakawa.common.ar_types import NPath
from arakawa.exceptions import ARError
from arakawa.view import Blocks, BlocksT

from .file_store import B64FileEntry, GzipTmpFileEntry
from .processors import (
    ConvertXML,
    ExportHTMLFileAssets,
    ExportHTMLInlineAssets,
    ExportHTMLStringInlineAssets,
    PreProcessView,
)
from .types import Formatting, Pipeline, ViewState


################################################################################
# exported public API
def build_report(
    blocks: BlocksT,
    name: str | None = None,
    dest: NPath | None = None,
    formatting: Formatting | None = None,
    cdn_base: str | None = None,
    overwrite: bool = False,
) -> None:
    """Build an (static) app with a directory structure, which can be served by a local http server

    !!! note
        This outputs compressed assets into the dir as well, may be an issue if self-hosting

    Args:
        blocks: The `Blocks` object or a list of Blocks
        name: The name of the app directory to be created
        dest: File path to store the app directory
        formatting: Sets the basic app styling
        overwrite: Replace existing app with the same name and destination if already exists (default: False)
    """
    name = name or "Report"

    # build the dest dir
    app_dir: Path = Path(dest or os.getcwd()) / name
    app_exists = app_dir.is_dir()

    if app_exists and overwrite:
        rmtree(app_dir)
    elif app_exists and not overwrite:
        raise ARError(
            f"Report exists at given path {app_dir!s} -- set `overwrite=True` to allow overwrite"
        )

    assets_dir = app_dir / "assets"
    assets_dir.mkdir(parents=True)

    # write the app html and assets
    s = ViewState(
        blocks=Blocks.wrap_blocks(blocks),
        file_entry_klass=GzipTmpFileEntry,
        dir_path=assets_dir,
    )
    _: str = (
        Pipeline(s)
        .pipe(PreProcessView(is_finalized=True))
        .pipe(ConvertXML())
        .pipe(
            ExportHTMLFileAssets(
                app_dir=app_dir,
                name=name,
                formatting=formatting,
                cdn_base=cdn_base,
            )
        )
        .result
    )


def save_report(
    blocks: BlocksT,
    path: str,
    open: bool = False,
    name: str | None = None,
    formatting: Formatting | None = None,
    cdn_base: str | None = None,
) -> None:
    """Save the app document to a local HTML file

    Args:
        blocks: The `Blocks` object or a list of Blocks
        path: File path to store the document
        open: Open in your browser after creating (default: False)
        name: Name of the document (optional: uses path if not provided)
        formatting: Sets the basic app styling
    """
    s = ViewState(blocks=Blocks.wrap_blocks(blocks), file_entry_klass=B64FileEntry)
    _: str = (
        Pipeline(s)
        .pipe(PreProcessView(is_finalized=True))
        .pipe(ConvertXML())
        .pipe(
            ExportHTMLInlineAssets(
                path=path,
                open=open,
                name=name or "Report",
                formatting=formatting,
                cdn_base=cdn_base,
            )
        )
        .result
    )


def stringify_report(
    blocks: BlocksT,
    name: str | None = None,
    formatting: Formatting | None = None,
    cdn_base: str | None = None,
) -> str:
    """Stringify the app document to a HTML string

    Args:
        blocks: The `Blocks` object or a list of Blocks
        name: Name of the document (optional: uses path if not provided)
        formatting: Sets the basic app styling
    """
    s = ViewState(blocks=Blocks.wrap_blocks(blocks), file_entry_klass=B64FileEntry)
    report_html: str = (
        Pipeline(s)
        .pipe(PreProcessView(is_finalized=False))
        .pipe(ConvertXML())
        .pipe(
            ExportHTMLStringInlineAssets(
                name=name or "Report",
                formatting=formatting,
                cdn_base=cdn_base,
            )
        )
        .result
    )

    return report_html



"""
Dynamic handling for optional libraries - this module is imported on demand
"""

from __future__ import annotations

from packaging import version as v
from packaging.specifiers import SpecifierSet

from arakawa.client import log

# NOTE - need to update this and keep in sync with JS
BOKEH_V_SPECIFIER = SpecifierSet(">=3.4.0,<3.5.0")
PLOTLY_V_SPECIFIER = SpecifierSet(">=5.24.1")
FOLIUM_V_SPECIFIER = SpecifierSet(">=0.17.0")


def _check_version(name: str, _v: v.Version, ss: SpecifierSet):
    if _v not in ss:
        log.warning(
            f"{name} version {_v} is not supported, these plots may not display correctly, please install version {ss}"
        )


# Optional Plotting library import handling
# Matplotlib
try:
    from matplotlib.axes import Axes  # noqa: F401
    from matplotlib.figure import Figure  # noqa: F401
    from numpy import ndarray  # noqa: F401

    HAVE_MATPLOTLIB = True
except ImportError:
    log.debug("No matplotlib found")
    HAVE_MATPLOTLIB = False

# Folium
try:
    import folium
    from folium import Map  # noqa: F401

    _check_version("Folium", v.Version(folium.__version__), FOLIUM_V_SPECIFIER)
    HAVE_FOLIUM = True
except ImportError:
    HAVE_FOLIUM = False
    log.debug("No folium found")

# Bokeh
try:
    import bokeh
    from bokeh.models import LayoutDOM as BLayout  # noqa: F401
    from bokeh.plotting import figure as BFigure  # noqa: F401, N812

    _check_version("Bokeh", v.Version(bokeh.__version__), BOKEH_V_SPECIFIER)
    HAVE_BOKEH = True
except ImportError:
    HAVE_BOKEH = False
    log.debug("No Bokeh Found")

# Plotly
try:
    import plotly
    from plotly.graph_objects import Figure as PFigure  # noqa: F401

    _check_version("Plotly", v.Version(plotly.__version__), PLOTLY_V_SPECIFIER)
    HAVE_PLOTLY = True
except ImportError:
    HAVE_PLOTLY = False
    log.debug("No Plotly Found")



"""Type-driven block wrapping"""
# ruff: noqa: FA100

from pathlib import Path
from typing import Union

import pandas as pd
from altair.utils import SchemaBase
from multimethod import multimethod

from arakawa import blocks as b
from arakawa import optional_libs as opt
from arakawa.exceptions import ARError

from .base import DataBlock


@multimethod
def convert_to_block(x: object) -> DataBlock:
    raise ARError(
        f"{type(x)} not supported directly, please pass into in the appropriate ar object (including ar.Attachment if want to upload as a pickle)"
    )


# NOTE - this is currently disabled to avoid confusing users when they
# try to embed any Python object, instead they must use ar.Attachment
# @multimethod
# def convert_to_block(x: Any) -> DataBlock:
#     return b.Attachment(x)


@multimethod
def convert_to_block(x: str) -> DataBlock:
    return b.Text(x)


@multimethod
def convert_to_block(x: Path) -> DataBlock:
    return b.Attachment(file=x)


@multimethod
def convert_to_block(x: pd.DataFrame) -> DataBlock:
    n_cells = x.shape[0] * x.shape[1]
    return b.Table(x) if n_cells <= 250 else b.DataTable(x)


# Plots
@multimethod
def convert_to_block(x: SchemaBase) -> DataBlock:
    return b.Plot(x)


if opt.HAVE_BOKEH:

    @multimethod
    def convert_to_block(x: Union[opt.BFigure, opt.BLayout]) -> DataBlock:
        return b.Plot(x)


if opt.HAVE_PLOTLY:

    @multimethod
    def convert_to_block(x: opt.PFigure) -> DataBlock:
        return b.Plot(x)


if opt.HAVE_FOLIUM:

    @multimethod
    def convert_to_block(x: opt.Map) -> DataBlock:
        return b.Plot(x)


if opt.HAVE_MATPLOTLIB:

    @multimethod
    def convert_to_block(x: Union[opt.Figure, opt.Axes, opt.ndarray]) -> DataBlock:
        return b.Plot(x)



from __future__ import annotations

import datetime
import gzip
import io
import os
import shutil
import subprocess
import time
from collections.abc import Generator, Iterable
from contextlib import contextmanager
from pathlib import Path
from tempfile import (
    NamedTemporaryFile,
    TemporaryDirectory,
    _TemporaryFileWrapper,
    mkstemp,
)
from typing import BinaryIO

from loguru import logger as log

from .ar_types import NPath
from .utils import ON_WINDOWS


@contextmanager
def log_command(command: str) -> Generator[None]:
    """Log an internal process"""
    log.info(f"Starting {command}")
    yield
    log.info(f"Finished {command}")


@contextmanager
def create_temp_file(
    suffix: str, prefix: str = "arakawa-temp-", mode: str = "w+b"
) -> Generator[_TemporaryFileWrapper]:
    """Creates a NamedTemporaryFile that doesn't disappear on .close()"""
    temp_file = NamedTemporaryFile(
        suffix=suffix, prefix=prefix, mode=mode, delete=False
    )
    try:
        yield temp_file
    finally:
        os.unlink(temp_file.name)


@contextmanager
def temp_fname(
    suffix: str, prefix: str = "arakawa-temp-", keep: bool = False
) -> Generator[str]:
    """Wrapper to generate a temporary filename only that is deleted on leaving context"""
    # TODO - return Path
    (in_f, in_f_name) = mkstemp(suffix=suffix, prefix=prefix)
    try:
        os.close(in_f)
        yield in_f_name
    finally:
        if os.path.exists(in_f_name) and not keep:
            os.unlink(in_f_name)


@contextmanager
def unix_compress_file(f_name: NPath, level: int = 6) -> Generator[str]:
    """(UNIX only) Return path to a compressed version of the input filename"""
    subprocess.run(["gzip", "-kf", f"-{level}", f_name], check=True)
    f_name_gz = f"{f_name}.gz"
    try:
        yield f_name_gz
    finally:
        os.unlink(f_name_gz)


@contextmanager
def unix_decompress_file(f_name: NPath) -> Generator[str]:
    """(UNIX only) Return path to a compressed version of the input filename"""
    subprocess.run(["gunzip", "-kf", f_name], check=True)
    f_name_gz = f"{f_name}.gz"
    try:
        yield f_name_gz
    finally:
        os.unlink(f_name_gz)


@contextmanager
def compress_file(f_name: NPath, level: int = 6) -> Generator[str]:
    """(X-Plat) Return path to a compressed version of the input filename"""
    f_name_gz = f"{f_name}.gz"
    with (
        open(f_name, "rb") as f_in,
        gzip.open(f_name_gz, "wb", compresslevel=level) as f_out,
    ):
        shutil.copyfileobj(f_in, f_out)
    try:
        yield f_name_gz
    finally:
        # NOTE - disable on windows temporarily
        if not ON_WINDOWS:
            os.unlink(f_name_gz)


def inmemory_compress(content: BinaryIO) -> BinaryIO:
    """(x-plat) Memory-based gzip compression"""
    content.seek(0)
    zbuf = io.BytesIO()
    with gzip.GzipFile(mode="wb", fileobj=zbuf, mtime=0.0) as zfile:
        zfile.write(content.read())
    zbuf.seek(0)
    return zbuf


@contextmanager
def temp_workdir() -> Generator[None]:
    """Set working dir to a tempdir for duration of context"""
    with TemporaryDirectory() as tmp_dir:
        curdir = os.getcwd()
        os.chdir(tmp_dir)
        try:
            yield None
        finally:
            os.chdir(curdir)


@contextmanager
def pushd(
    directory: NPath, pre_create: bool = False, post_remove: bool = False
) -> Generator[None]:
    """Switch dir and push it onto the (call-)stack"""
    directory = Path(directory)
    cwd = os.getcwd()
    log.debug(f"[cd] {cwd} -> {directory}")
    if not directory.exists() and pre_create:
        Path(directory).mkdir(parents=True)
    os.chdir(directory)
    try:
        yield
    finally:
        log.debug(f"[cd] {cwd} <- {directory}")
        os.chdir(cwd)
        if post_remove:
            shutil.rmtree(directory, ignore_errors=True)


def get_filesize(filename: Path) -> int:
    return filename.stat().st_size


def walk_path(path: Path) -> Iterable[Path]:
    for p in path.rglob("*"):
        if not p.is_dir():
            yield p


def unixtime() -> int:
    return int(time.time())


def timestamp(x: datetime.datetime | None = None) -> str:
    """Return ISO timestamp for a datetime"""
    x = x or datetime.datetime.utcnow()
    return f'{x.isoformat(timespec="seconds")}{"" if x.tzinfo else "Z"}'



"""
Arakawa built-in helper functions to make creating your reports a bit simpler and reduce common tasks
"""

from __future__ import annotations

from pathlib import Path
from typing import cast

from . import Blocks
from . import blocks as b
from .common import NPath


def add_code(
    block: b.BlockOrPrimitive, code: str, language: str = "python"
) -> b.Select:
    """
    Attach code fragment to an existing plot/figure/dataframe for use within a report

    Args:
        block: The existing object to add code to - can be either an existing ar Block or an Python object
        code: The code fragment to add
        language: The language of the code fragment (optional)

    Returns:
        A Select block that provides the figure and the code in tabs that can be selected by the user
    """

    w_block = b.wrap_block(block)
    w_block._add_attributes(label="Figure")
    return b.Select(
        w_block, b.Code(code, language, label="Code"), type=b.SelectType.TABS
    )


def build_md_view(
    text_or_file: str | NPath,
    *args: b.BlockOrPrimitive,
    **kwargs: b.BlockOrPrimitive,
) -> Blocks:
    """
    An easy way to build a complete report from a single top-level markdown text / file template.
    Any additional context can be passed in and will be inserted into the Markdown template.

    Args:
        text_or_file: The markdown text, or path to a markdown file, using {{}} for templating
        *args: positional template context arguments
        **kwargs: keyword template context arguments

    Returns:
        An Arakawa App object for saving or uploading

    ..tip:: Either text or file is required as input
    ..tip:: Context, via args/kwargs can be plain Python objects, e.g. dataframes, and plots, or Arakawa blocks, e.g. ar.Plot, etc.

    """
    try:
        b_text = (
            b.Text(file=text_or_file)
            if Path(text_or_file).exists()
            else b.Text(text=cast(str, text_or_file))
        )
    except OSError:
        b_text = b.Text(text=cast(str, text_or_file))

    group = b_text.format(*args, **kwargs)
    return Blocks(group)



from __future__ import annotations

from collections.abc import Mapping
from copy import copy
from typing import TYPE_CHECKING, Union

from lxml import etree
from lxml.etree import _Element as ElementT
from typing_extensions import Self

from arakawa.blocks import Group
from arakawa.blocks.base import BlockOrPrimitive
from arakawa.blocks.layout import ContainerBlock

if TYPE_CHECKING:
    from arakawa.processors.types import Formatting


class Blocks(ContainerBlock):
    """Container that holds a collection of blocks"""

    # This is essentially an easy-to-use wrapper around a list of blocks
    # that is composable.
    # TODO - move to arakawa.blocks ?

    _tag = "Blocks"

    def __init__(
        self,
        *arg_blocks: BlockOrPrimitive,
        blocks: list[BlockOrPrimitive] | None = None,
        **kwargs,
    ):
        # if passed a single View into a View object, pull out the contained blocks and use instead
        if len(arg_blocks) == 1 and isinstance(arg_blocks[0], Blocks):
            arg_blocks = tuple(arg_blocks[0].blocks)

        super().__init__(*arg_blocks, blocks=blocks, **kwargs)

    def __or__(self, other: Self):
        x = Group(blocks=self.blocks) if len(self.blocks) > 1 else self.blocks[0]
        y = Group(blocks=other.blocks) if len(other.blocks) > 1 else other.blocks[0]
        z = Group(x, y, columns=2)
        return Blocks(z)

    @classmethod
    def from_notebook(
        cls,
        opt_out: bool = True,
        show_code: bool = False,
        show_markdown: bool = True,
        template: str = "auto",
    ) -> Self:
        from arakawa.ipython import templates as ip_t
        from arakawa.ipython.utils import cells_to_blocks

        blocks = cells_to_blocks(
            opt_out=opt_out, show_code=show_code, show_markdown=show_markdown
        )
        app_template_cls = ip_t._registry.get(template) or ip_t.guess_template(blocks)
        app_template = app_template_cls(blocks)
        app_template.transform()
        app_template.validate()
        return cls(blocks=app_template.blocks)

    def get_dom(self) -> ElementT:
        """Return the Document structure for the View"""
        # internal debugging method
        from arakawa.processors.file_store import DummyFileEntry, FileStore

        from .xml_visitor import XMLBuilder

        builder = XMLBuilder(FileStore(DummyFileEntry))
        self.accept(builder)
        return builder.get_root()

    def get_dom_str(self) -> str:
        dom = self.get_dom()
        return etree.tounicode(dom, pretty_print=True)

    def pprint(self) -> None:
        from .visitors import PrettyPrinter

        self.accept(PrettyPrinter())

    @classmethod
    def wrap_blocks(cls, x: Self | list[BlockOrPrimitive] | BlockOrPrimitive) -> Self:
        blocks: Self
        if isinstance(x, Blocks):
            blocks = copy(x)
        elif isinstance(x, list):
            blocks = cls(*x)
        else:
            blocks = cls(x)
        return blocks

    @property
    def has_compute(self):
        return False


class View(Blocks):
    pass


BlocksT = Union[
    Blocks,
    list[BlockOrPrimitive],
    Mapping[str, BlockOrPrimitive],
    BlockOrPrimitive,
]


class Report(Blocks):
    """
    App documents collate plots, text, tables, and files into an interactive document that
    can be analysed and shared by users in their browser
    """

    # Backwards compatible interfaces/wrappers

    def __init__(
        self,
        *arg_blocks: BlockOrPrimitive,
        blocks: list[BlockOrPrimitive] | None = None,
        **kwargs,
    ):
        super().__init__(*arg_blocks, blocks=blocks, **kwargs)

    def save(
        self,
        path: str,
        open: bool = False,
        name: str | None = None,
        formatting: Formatting | None = None,
        cdn_base: str | None = None,
    ) -> None:
        from ..processors import save_report

        save_report(
            blocks=self,
            path=path,
            open=open,
            name=name,
            formatting=formatting,
            cdn_base=cdn_base,
        )

    def stringify(
        self,
        name: str | None = None,
        formatting: Formatting | None = None,
        cdn_base: str | None = None,
        template_name: str = "template.html",
    ) -> str:
        from ..processors import stringify_report

        return stringify_report(
            blocks=self,
            name=name,
            formatting=formatting,
            cdn_base=cdn_base,
        )
